import { Button, Callout, Collapse, InputGroup, Intent } from '@blueprintjs/core';
import { ArrowRight16, Password16, Warning20 } from '@carbon/icons-react';
import { css } from '@emotion/react';
import { CarbonIcon, UsernameInput } from '@redeye/client/components';
import { useStore } from '@redeye/client/store';
import { routes } from '@redeye/client/store/routing/router';
import { Views } from '@redeye/client/types';
import { useQuery } from '@tanstack/react-query';
import { observer } from 'mobx-react-lite';
import type { ComponentProps, FormEvent } from 'react';
import { createState } from '../mobx-create-state';

type LoginFormProps = ComponentProps<'form'> & {
	submitText?: string;
};

const isDevelop = import.meta.env.DEV;

export const LoginForm = observer<LoginFormProps>(({ onSubmit, submitText = 'Login', ...props }) => {
	const store = useStore();
	const { data, refetch } = useQuery(['users'], async () => await store.graphqlStore.queryGlobalOperators({}));
	const state = createState({
		username: isDevelop ? store.auth.userName || 'dev' : store.auth.userName || '',
		password: '',
		loading: false,
		errorMessage: '',
		*handleSubmit(event: FormEvent<HTMLFormElement>) {
			event.preventDefault();
			this.loading = true;
			if (store.appMeta.blueTeam) {
				store.auth.setUser(this.username);
				store.router.updateRoute({ path: routes[Views.CAMPAIGNS_LIST], params: { id: 'all' } });
				if (onSubmit) onSubmit(event);
			} else {
				// Make login call
				const formData = new FormData();
				formData.append('password', this.password);
				try {
					const loginResponse: Response = yield fetch(`${store.auth.serverUrl}/api/login`, {
						method: 'POST',
						mode: 'cors',
						cache: 'no-cache',
						credentials: 'include',
						body: formData,
					});
					// TODO: could set a state that makes the password and server inputs invalid?
					this.loading = false;
					if (loginResponse.status === 401) {
						if (!this.password) this.errorMessage = 'Password Required';
						else this.errorMessage = 'Incorrect password';
					} else if (loginResponse.status === 400 || loginResponse.status > 401)
						this.errorMessage = 'Error communicating with server';
					else if (loginResponse.status !== 200) this.errorMessage = 'Error logging in';
					else {
						this.errorMessage = '';
						store.auth.setUser(this.username);
						store.router.updateRoute({ path: routes[Views.CAMPAIGNS_LIST], params: { id: 'all' } });
						if (onSubmit) onSubmit(event);
					}
				} catch (e) {
					this.loading = false;
					this.errorMessage = 'Error communicating with server';
				}
			}
		},
	});

	return (
		<form {...props} onSubmit={state.handleSubmit} autoComplete="on">
			{!store.appMeta.blueTeam && (
				<>
					<InputGroup
						value={state.password}
						onChange={(e) => state.update('password', e.target.value)}
						cy-test="password"
						autoComplete="password"
						type="password"
						name="password"
						placeholder="server password"
						css={inputSpacingTightStyle}
						leftIcon={<CarbonIcon icon={Password16} />}
						large
					/>
				</>
			)}
			<UsernameInput
				cy-test="username"
				username={state.username}
				refetch={refetch}
				users={data?.globalOperators}
				updateUser={(userName) => state.update('username', userName)}
				css={otherSpacingLooseStyle}
			/>
			<Collapse isOpen={!!state.errorMessage}>
				<Callout
					css={otherSpacingLooseStyle}
					intent={Intent.DANGER}
					icon={<CarbonIcon icon={Warning20} />}
					children={state.errorMessage}
				/>
			</Collapse>
			<Button
				cy-test="login-btn"
				text={submitText}
				loading={state.loading}
				intent="primary"
				css={otherSpacingLooseStyle}
				disabled={state.username.length < 1}
				type="submit"
				rightIcon={<CarbonIcon icon={ArrowRight16} />}
				large
			/>
		</form>
	);
});

const inputSpacingTightStyle = css`
	margin-bottom: 0.5rem;
`;
const otherSpacingLooseStyle = css`
	margin-bottom: 1rem;
`;
// const removeAutofillStyle = css`
/* // don't think this is a good idea
box-shadow: rgba(255, 255, 255, 0.4) 0px -1px 0px 0px inset !important;
background: rgba(255, 255, 255, 0.06);
input:not(:-webkit-autofill) {
  background: transparent !important;
}
input:-webkit-autofill {
  -webkit-text-fill-color: rgba(255, 255, 255, 0.96);
  -webkit-background-image: none !important;
  -webkit-background-clip: text;
}
*/
// `;
